#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>
#include <framebuffer.h>
#include <terminal.h>
#include <limine.h>
#include <hcf.h>
#include <memory.h>

LIMINE_BASE_REVISION(1)

struct limine_memmap_request memmap_request = {
	.id = LIMINE_MEMMAP_REQUEST,
	.revision = 0
};

void* memcpy(void* dest, const void* src, size_t n) {
	uint8_t* pdest = (uint8_t*)dest;
	const uint8_t* psrc = (const uint8_t*)src;

	for(size_t i = 0; i < n; i++) {
		pdest[i] = psrc[i];
	}

	return dest;
}

void* memset(void* s, int c, size_t n) {
	uint8_t* p = (uint8_t*)s;

	for(size_t i = 0; i < n; i++) {
		p[i] = (uint8_t)c;
	}

	return s;
}

void* memmove(void* dest, const void* src, size_t n) {
	uint8_t* pdest = (uint8_t*)dest;
	const uint8_t* psrc = (const uint8_t*)src;

	if(src > dest) {
		for(size_t i = 0; i < n; i++) {
			pdest[i] = psrc[i];
		}
	} else if(src < dest) {
		for(size_t i = n; i > 0; i--) {
			pdest[i-1] = psrc[i-1];
		}
	}

	return dest;
}

int memcmp(const void* s1, const void* s2, size_t n) {
	const uint8_t* p1 = (const uint8_t*)s1;
	const uint8_t* p2 = (const uint8_t*)s2;

	for(size_t i = 0; i < n; i++) {
		if(p1[i] != p2[i]) {
			return p1[i] < p2[i] ? -1 : 1;
		}
	}

	return 0;
}

const char* types[] = {
	"USABLE",
	"RESERVED",
	"ACPI_RECLAIMABLE",
	"ACPI_NVS",
	"BAD_MEMORY",
	"BOOTLOADER_RECLAIMABLE",
	"KERNEL_AND_MODULES",
	"FRAMEBUFFER"
};

void _start(void) {
	if(LIMINE_BASE_REVISION_SUPPORTED == false) {
		hcf();
	}
	
	fb_init();
	term_init();
	mem_init();

	printf("running at %dx%d\n", framebuffer->width, framebuffer->height);

	struct limine_memmap_response* memmap = memmap_request.response;	

	printf("%d memmap entries present.\n", memmap->entry_count);

	for(uint64_t i = 0; i < memmap->entry_count; i++) {
		struct limine_memmap_entry* entry = memmap->entries[i];
		if(entry->type == LIMINE_MEMMAP_USABLE)
			mem_add(entry->base, entry->length);

		printf("0x%010x - 0x%010x (type: % 22s, size: % 10u)\n", 
			entry->base, entry->base + entry->length, types[entry->type], entry->length);
	}

	void* a = mem_alloc(10);
	printf("memory allocated at %x", a); 

	hcf();
}
